#
# Authors: Rafael Ponce (1)
#          Miguel Cardenas-Montes (2)
#          Ignacio Sevilla (3)
#
# (1): rafael.ponce@ciemat.es
# (2): miguel.cardenas@ciemat.es
# (3): ignacio.sevilla@ciemat.es
#
# NOTE: This code has been successfully tested on NVIDIA GT555M. This is a gaming GPU, for this reason does not accept too many points
#
#
#                                               TO RUN
#
# This program calculates the 2pt-angular auto-correlation and cross-correlation functions by brute froce using a GPU
# The program receives as input parameters the following:
#
#    Catalog_1: The path of your galaxy catalog, usually is a galaxy catalog for a specific redshift bin
#    Catalog_2: The path of the same galaxy catalog you put as the first input (if you want to do the auto-correlation function) or another galaxy catalog (if you want to do the cross-correlation function between two redshift bins) 
#    Random_catalog: The path corresponding to the random catalog
#    Number_of_points: It's the number of bins per degree that you want
#    Output_correlation_file: The path of your output file 
#
# And the program will return inside the output file six or seven columns, depend on you did the auto-correlation or the cross-correlation function.
#
# The input files must only have two columns:
#
#    RA: Right Ascension in degrees 
#    DEC: Declination also in degrees
#
# The output will be:
#
#    For the auto-correlation:
#
#    angle(degrees) 
#    W(corr-func.) 
#    Poissonian Error 
#    Real-Real Couples 
#    Real-Rand Couples 
#    Rand-Rand Couples
#
#    For the cross-correlation:
#
#    angle(degrees) 
#    W(corr-func.) 
#    Poissonian Error 
#    Real_1-Real_2 Couples 
#    Real_1-Rand Couples 
#    Real_2-Rand Couples
#    Rand-Rand Couples
#
# To run the program:
#
# If you want to do calculate the auto-correlation you must do:
#
#        python GP2PCF_GPU_v0.2.py Catalog_1.txt Catalog_1.txt Random_cat.txt points output.txt
#
# If you want to do an cross-correlation you must do:
#
#        python GP2PCF_GPU_v0.2.py Catalog_1.txt Catalog_2.txt Random_cat.txt points output.txt
#
# EXAMPLES TO CALL THE FUNCTION:
#
#   To do the correlation with 16 points per degree
#
#       python GP2PCF_GPU_v0.2.py cat_1.txt cat_1.txt cat_rand.txt 16 W_auto_output.txt
#
#   To do cross-correlation with 32 points per degree
#
#       python GP2PCF_GPU_v0.2.py cat_1.txt cat_2.txt cat_rand.txt 32 W_cross_output.txt    
#
#
# IMPORTANT: 
#
# 1.- Note that if the first and second file have the same name, the program will calculate the auto-correlation function. If not, the program will calculate the cross-correlation function
#
# 2.- For memory reasons the points per degree should be between 4 points and 64 points and it must be multiple of 4.
#
# 3.- For hardware limitations, there is a relationship between the points per degree and the degree you can reach. You can see this relationship in the next table
#
# 4.- To install PyCuda please visit http://wiki.tiker.net/WelcomePage
#
# Points	Angle
# 04	    64
# 08		32
# 16		16
# 32		08
# 64		04
#



# Libraries

import pycuda.driver as cuda
import pycuda.autoinit
from pycuda.compiler import SourceModule
import time

import numpy as N
import math as M
import sys 

# This function counts the number of lines in a input file   

def counting_lines(input_file):
    lines=0
    f_in=open(input_file,'r')
    for line in f_in.xreadlines():
        lines=lines+1
    f_in.close()
    return(lines)

# This function checks if the file exists or not

def file_exist(file_name):
    try:
        file_name = open(file_name,'r')
        file_name.close()
        exist=1
    except:
        exist=0
    return(exist)

# This function checks the number of columns

def check_columns(file_name):
    f_in=open(file_name,'r')
    line=f_in.readline()
    columns=len(line.split())
    if (columns!=2):
        return(0)
    else:
        return(1)

# This function verify the input parameters

def verification(input_parameters):

    # Checking if any input parameter is missing

    if len(input_parameters)!=6:
        print 'You must exactly introduce 5 parameters: catalog_1 catalog_2 random_catalog points_per_degree output_file'
        exit(0)
    f_first_cat=input_parameters[1]
    f_second_cat=input_parameters[2]
    f_random_cat=input_parameters[3] 
    points=float(input_parameters[4])
    f_out=input_parameters[5]

    # Checking if the input files exist or not

    if (file_exist(f_first_cat)==0):
        print 'Error opening data file number 1'
        exit(0)
    if (file_exist(f_second_cat)==0):
        print 'Error opening data file number 2'
        exit(0)
    if (file_exist(f_random_cat)==0):
        print 'Error opening data file number 3'
        exit(0)

    # Checking the points per degree you want to

    if points<4 or points>64 or not(points%4==0):
        print 'The number of points per degree must be multiple of four and be between 4 and 64'
        exit(0)

    # Checking columns
    
    if(check_columns(f_first_cat)==0):
        print 'Number of columns in the FIRST file must be exactly 2 and the first one has to be the right ascention and the second one the declination, both in degrees'
        exit(0)

    if(check_columns(f_second_cat)==0):
        print 'Number of columns in the SECOND file must be exactly 2 and the first one has to be the right ascention and the second one the declination, both in degrees'
        exit(0)

    if(check_columns(f_random_cat)==0):
        print 'Number of columns in the THIRD (RANDOM) file must be exactly 2 and the first one has to be the right ascention and the second one the declination, both in degrees'
        exit(0)
 
    return(1)

# This function lets the program knows if you want to do a cross-correlation or a auto-correlation function 

# This function checks if the name of the first two input files are the same. If the name is the same for both, 
# the program makes the auto-correlation, if not, the program makes the cross-correlation

def cross_auto(f_first_cat,f_second_cat):
    if (f_first_cat==f_second_cat):
        return(0) # auto
    else:
        return(1) # cross

# Equatorial to cartesian conversion  

def eq2cart(f_cat,xd,yd,zd):
    f_in=open(f_cat,'r')
    for i,line in enumerate(f_in.xreadlines()):
        # reading ecuatorial coordinates
        a=line.split()
        phi=float(a[0])*N.pi/180
        cth=M.cos((90-float(a[1]))*N.pi/180)
        th=M.acos(cth)
        
        # to cartesian coordinatesa
        
        xd[i]=M.cos(phi)*M.sin(th)
        yd[i]=M.sin(phi)*M.sin(th)
        zd[i]=cth
    f_in.close()    
    return(xd,yd,zd)


def crosscorr():

    # Taking input parameters

    input_parameters=sys.argv

    # Checking if the input files and call to script meet the requirements

    if (verification(input_parameters)==0):
        exit(1)

    # Assignment of Variables with inputs    

    f_first_cat=input_parameters[1]
    f_second_cat=input_parameters[2]
    f_random_cat=input_parameters[3] 
    points_per_degree=float(input_parameters[4])
    f_corr=input_parameters[5]

    # Defining what angle we can reach

    size_block=256 # Depend on the GPU architecture
    max_angle=float(size_block)/float(points_per_degree)

    # Kernels

    # This kernel counts the number of couples that there are between a data with themself
    # We will use this kernel to calculate real-real couples and random-random couples

    kernel_1 = SourceModule("""
    __global__ void binning(float *xd,float *yd,float *zd,float *ZZ,int number_lines,int points_per_degree, int number_of_degrees)
    {

        /* We define variables (arrays) in shared memory */

        __shared__ float angle["""+str(size_block)+"""];
        __shared__ float temp["""+str(size_block)+"""];

        /* We define an index to run through these two arrays */    

        int index = threadIdx.x;

        /* This variable is necesary to accelerate the calculation, it's due that "temp" was definied in the shared memory too */

        temp[index]=0;
        float x,y,z; //MCM
        float xx,yy,zz; //MCM

        /* We start the counting */

        for (int i=0;i<number_lines;i++)
        {
            int dim_idx =  blockIdx.x * blockDim.x + threadIdx.x;
            x = xd[i];//MCM
            y = yd[i];//MCM
            z = zd[i];//MCM

            /* The "while" replaces the second for-loop in the sequential calculation case (CPU). We use "while" rather than "if" as recommended in the book "Cuda by Example" */

            while (dim_idx < number_lines)
            {
                xx = xd[dim_idx];//MCM
                yy = yd[dim_idx];//MCM
                zz = zd[dim_idx];//MCM

                /* We make the dot product */ 
                angle[index] = x * xx + y * yy + z * zz;//MCM

                /* Sometimes "angle" is higher than one, due to numnerical precision, to solve it we use the next sentence */

                angle[index]=fminf(angle[index],1.0);
                angle[index]=acosf(angle[index])*180.0/M_PI;
            
                /* We finally count the number of pairs separated an angular distance "angle", always in shared memory */

                if(angle[index] < """+str(max_angle)+""")
                {
                    atomicAdd( &temp[int(angle[index]*"""+str(points_per_degree)+""")], 1.0);
                }
                __syncthreads();
                dim_idx = blockDim.x * gridDim.x+dim_idx;
            }
        }

        /* We copy the number of pairs from shared memory to global memory */

        atomicAdd( &ZZ[threadIdx.x] , temp[threadIdx.x]);
        __syncthreads();
    }
    """)


    # This kernel counts the number of couples that there are between two data groups
    # We will use this kernel to calculate real-random couples and real_1-real_2 couples

    kernel_2 = SourceModule("""
    __global__ void binning_mix(float *xd_real, float *yd_real, float *zd_real, float *xd_sim, float *yd_sim, float *zd_sim, float *ZY, int lines_number_1, int lines_number_2, int points_per_degree, int number_of_degrees)
        {

        /* We define variables (arrays) in shared memory */

        __shared__ float angle["""+str(size_block)+"""];
        __shared__ float temp["""+str(size_block)+"""];

        /* We define an index to run through these two arrays */    

        int index = threadIdx.x;

        /* This variable is necesary to accelerate the calculation, it's due that "temp" was definied in the shared memory too */

        temp[index]=0;
        float x,y,z; //MCM
        float xx,yy,zz; //MCM

        /* We start the counting */

        for (int i=0;i<lines_number_1;i++)
        {
            int dim_idx =  blockIdx.x * blockDim.x + threadIdx.x;
            x = xd_real[i];//MCM
            y = yd_real[i];//MCM
            z = zd_real[i];//MCM

            /* The "while" replaces the second for-loop in the sequential calculation case (CPU). We use "while" rather than "if" as recommended in the book "Cuda by Example" */

            while (dim_idx < lines_number_2)
            {
                xx = xd_sim[dim_idx];//MCM
                yy = yd_sim[dim_idx];//MCM
                zz = zd_sim[dim_idx];//MCM
                /* We make the dot product */ 
                angle[index] = x * xx + y * yy + z * zz;//MCM

                /* Sometimes "angle" is higher than one, due to numnerical precision, to solve it we use the next sentence */
            
                angle[index]=fminf(angle[index],1.0);
                angle[index]=acosf(angle[index])*180.0/M_PI;
            
                /* We finally count the number of pairs separated an angular distance "angle", always in shared memory */

                if(angle[index] < """+str(max_angle)+""")
                {
                    atomicAdd( &temp[int(angle[index]*"""+str(points_per_degree)+""")], 1.0);
                }
                __syncthreads();
                dim_idx = blockDim.x * gridDim.x+dim_idx;
            }
        }

        /* We copy the number of pairs from shared memory to global memory */

        atomicAdd( &ZY[threadIdx.x] , temp[threadIdx.x]);
        __syncthreads();
    }
    """)




    # Opening the first input file

    f_real_1=open(f_first_cat,'r')

    # Counting lines

    real_lines_1=counting_lines(f_first_cat)

    # setting the GPU integer format

    real_lines_1=N.int32(real_lines_1)

    # We define variables to store the first real data

    xd_real_1=N.zeros(real_lines_1)
    yd_real_1=N.zeros(real_lines_1)
    zd_real_1=N.zeros(real_lines_1)

    # Opening the first input file
        
    (xd_real_1,yd_real_2,zd_real_1)=eq2cart(f_first_cat,xd_real_1,yd_real_1,zd_real_1)
        
    # setting data with GPU format
    
    xd_real_1=xd_real_1.astype(N.float32)
    yd_real_1=yd_real_1.astype(N.float32)
    zd_real_1=zd_real_1.astype(N.float32)
    
    # Closing the first input file
    
    f_real_1.close()

    if (cross_auto(f_first_cat,f_second_cat)==1):
        # Opening the second input file

        f_real_2=open(f_second_cat,'r')

        # Counting lines

        real_lines_2=counting_lines(f_second_cat)

        # setting the GPU integer format

        real_lines_2=N.int32(real_lines_2)

        # We define variables to store the second real data

        xd_real_2=N.zeros(real_lines_2)
        yd_real_2=N.zeros(real_lines_2)
        zd_real_2=N.zeros(real_lines_2)
        
        (xd_real_2,yd_real_2,zd_real_2)=eq2cart(f_second_cat,xd_real_2,yd_real_2,zd_real_2)

        # setting data with GPU format
    
        xd_real_2=xd_real_2.astype(N.float32)
        yd_real_2=yd_real_2.astype(N.float32)
        zd_real_2=zd_real_2.astype(N.float32)
    
        # Closing the first input file
    
        f_real_2.close()

    # Opening the random input file

    f_rand=open(f_random_cat,'r')

    # Counting lines

    random_lines=counting_lines(f_random_cat)

    # setting the GPU integer format

    random_lines=N.int32(random_lines)

    # We define variables to store the random data

    xd_rand=N.zeros(random_lines)
    yd_rand=N.zeros(random_lines)
    zd_rand=N.zeros(random_lines)

    (xd_rand,yd_rand_2,zd_rand)=eq2cart(f_random_cat,xd_rand,yd_rand,zd_rand)
    
    # setting data with GPU format
        
    xd_rand=xd_rand.astype(N.float32)
    yd_rand=yd_rand.astype(N.float32)
    zd_rand=zd_rand.astype(N.float32)

    # We determine which is the maximum number of lines
    
    if (cross_auto(f_first_cat,f_second_cat)==0):
    
        max_lines=max(real_lines_1,random_lines)       
    
    else:
    
        max_lines=max(real_lines_1,real_lines_2,random_lines)

    # We define the GPU-GRID size, it's really the number of blocks we are going to use on the GPU

    blocks_number=int(float(max_lines)/float(size_block))+1

    if (cross_auto(f_first_cat,f_second_cat)==0):
        
        # We initialize and format the couples variables on CPU
    
        DD=N.zeros(size_block)
        DD=DD.astype(N.float32)
        RR=N.zeros(size_block)
        RR=RR.astype(N.float32)
        DR=N.zeros(size_block)
        DR=DR.astype(N.float32)
        
    else:
        
        # We initialize and format the couples variables on CPU
    
        D1D2=N.zeros(size_block)
        D1D2=D1D2.astype(N.float32)
        D1R=N.zeros(size_block)
        D1R=D1R.astype(N.float32)
        D2R=N.zeros(size_block)
        D2R=D2R.astype(N.float32)
        RR=N.zeros(size_block)
        RR=RR.astype(N.float32)


    # Unlike the CUDA code, in PyCuda we have a way to call the kernel in which we needn't to define every variable in the GPU, the kernel calling define everything automatically and we needn't to copy and recover any data, this way do it too.

    if (cross_auto(f_first_cat,f_second_cat)==0):

        # We define the variable to call the kernel we need to calculate the number of couples that there are between a group of datas with themself
  
        calling_first_kernel = kernel_1.get_function ( "binning" )

        # We call the needed kernel to calculate the number of couples that there are between the real data with themself     

        calling_first_kernel (cuda.In (xd_real_1), cuda.In (yd_real_1), cuda.In (zd_real_1), cuda.InOut (DD), real_lines_1 ,block=(size_block,1,1), grid=(blocks_number,1))

        # The next line maybe it wouldn't be necessary but to avoid memory problems we prefer to define it again

        calling_first_kernel = kernel_1.get_function ( "binning" )

        # We call the last kernel to calculate the number of couples that there are between the random data with themself

        calling_first_kernel (cuda.In (xd_rand), cuda.In (yd_rand), cuda.In (zd_rand), cuda.InOut (RR), random_lines ,block=(size_block,1,1), grid=(blocks_number,1))

        # We define the variable to call the kernel we need to calculate the number of couples that there are between the first real data and the random data

        calling_second_kernel = kernel_2.get_function ( "binning_mix" )

        # We call the needed kernel to calculate the number of couples that there are between the first real data and the random data

        calling_second_kernel (cuda.In (xd_real_1), cuda.In (yd_real_1), cuda.In (zd_real_1), cuda.In (xd_rand), cuda.In (yd_rand), cuda.In (zd_rand), cuda.InOut (DR), real_lines_1, random_lines ,block=(size_block,1,1), grid=(blocks_number,1))

    else:
        
        # We define the variable to call the kernel we need to calculate the number of couples that there are between the first real data and the second data, the first real data and the random data and finally the second real data and the random data

        calling_kernel_2 = kernel_2.get_function ( "binning_mix" )

        # We call the kernel to calculate the number of couples that there are between the first real data and second real data

        calling_kernel_2 (cuda.In (xd_real_1), cuda.In (yd_real_1), cuda.In (zd_real_1), cuda.In (xd_real_2), cuda.In (yd_real_2), cuda.In (zd_real_2), cuda.InOut (D1D2), real_lines_1, real_lines_2 ,block=(size_block,1,1), grid=(blocks_number,1))

        # We define it again to avoid memory problems

        calling_kernel_2 = kernel_2.get_function ( "binning_mix" )

        # We call the kernel to calculate the number of couples that there are between the first real data and the random data

        calling_kernel_2 (cuda.In (xd_real_1), cuda.In (yd_real_1), cuda.In (zd_real_1), cuda.In (xd_rand), cuda.In (yd_rand), cuda.In (zd_rand), cuda.InOut (D1R), real_lines_1, random_lines ,block=(size_block,1,1), grid=(blocks_number,1))

        # We define it again to avoid memory problems

        calling_kernel_2 = kernel_2.get_function ( "binning_mix" )

        # We call the kernel to calculate the number of couples that there are between the second real data and the random data

        calling_kernel_2 (cuda.In (xd_real_2), cuda.In (yd_real_2), cuda.In (zd_real_2), cuda.In (xd_rand), cuda.In (yd_rand), cuda.In (zd_rand), cuda.InOut (D2R), real_lines_2, random_lines ,block=(size_block,1,1), grid=(blocks_number,1))


        # We define the variable to call the kernel we need to calculate the number of couples that there are between a every data with themself

        calling_kernel_1 = kernel_1.get_function ( "binning" )

        # We call the kernel to calculate the number of couples that there are between the random ddta with themself

        calling_kernel_1 (cuda.In (xd_rand), cuda.In (yd_rand), cuda.In (zd_rand), cuda.InOut (RR), random_lines ,block=(size_block,1,1), grid=(blocks_number,1))
        
        
    # Opening the output file
    
    f_out=open(f_corr,'w')

    # We define variables we need to do the correlation function or cross-correlation function

    W=[]
    theta=[]
    
    if (cross_auto(f_first_cat,f_second_cat)==0):
        norm_const=float(random_lines)/float(real_lines_1)
    else:
        norm_const_1=float(random_lines)/float(real_lines_1)
        norm_const_2=float(random_lines)/float(real_lines_2)
    
    if (cross_auto(f_first_cat,f_second_cat)==0):
        for i in range(len(DD)):
            # We are calculating the Landy & Szalay estimator

            W_aux=(M.pow(norm_const,2)*DD[i]-2*(norm_const)*DR[i])/RR[i]+1.0
            W.append(W_aux)

            # The angle corresponding to the W value

            theta_aux=(1.0/points_per_degree)/2+(i*(1.0/points_per_degree))
            theta.append(theta_aux)

            # Poissonian error
            
            poissonian_error=(1.0+W_aux)/M.sqrt(DD[i])

            # We writting the auto-correlation in the output file

            if(theta_aux<max_angle):

                f_out.write(str(theta_aux)+'\t'+str(W_aux)+'\t'+str(poissonian_error)+'\t'+str(DD[i])+'\t'+str(DR[i])+'\t'+str(RR[i])+'\n')

    else:
        
        for i in range(1,len(D1D2)):

            # We are calculating the Landy & Szalay estimator

            W_aux=1+norm_const_1*norm_const_2*(D1D2[i]/RR[i])-norm_const_1*(D1R[i]/RR[i])-norm_const_2*(D2R[i]/RR[i])
            W.append(W_aux)

            # The angle corresponding to the W value

            theta_aux=(1.0/points_per_degree)/2+(i*(1.0/points_per_degree))
            theta.append(theta_aux)

            # Poissonian error

            poissonian_error=(1.0+W_aux)/M.sqrt(D1D2[i])

            # We writting the cross-correlation in the output file

            if(theta_aux<max_angle):
                f_out.write(str(theta_aux)+'\t'+str(W_aux)+'\t'+str(poissonian_error)+'\t'+str(D1D2[i])+'\t'+str(D1R[i])+'\t'+str(D2R[i])+'\t'+str(RR[i])+'\n')
        
    # Closing output file

    f_out.close()
    return(0)
crosscorr()
    
